/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.client;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.Items;
import net.minecraft.inventory.Container;
import net.minecraft.item.ItemStack;
import net.minecraft.util.ResourceLocation;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;
import net.minecraftforge.fml.common.network.IGuiHandler;

import net.dries007.tfc.TerraFirmaCraft;
import net.dries007.tfc.api.recipes.knapping.KnappingRecipe;
import net.dries007.tfc.api.types.Rock;
import net.dries007.tfc.api.util.IRockObject;
import net.dries007.tfc.client.gui.*;
import net.dries007.tfc.objects.container.*;
import net.dries007.tfc.objects.items.ItemsTFC;
import net.dries007.tfc.objects.items.ceramics.ItemMold;
import net.dries007.tfc.objects.items.ceramics.ItemSmallVessel;
import net.dries007.tfc.objects.items.rock.ItemRock;
import net.dries007.tfc.objects.te.*;
import net.dries007.tfc.util.Helpers;

import static net.dries007.tfc.api.util.TFCConstants.MOD_ID;

public class TFCGuiHandler implements IGuiHandler
{
    private static final ResourceLocation SMALL_INVENTORY_BACKGROUND = new ResourceLocation(MOD_ID, "textures/gui/small_inventory.png");
    private static final ResourceLocation CLAY_TEXTURE = new ResourceLocation(MOD_ID, "textures/gui/knapping/clay_button.png");
    private static final ResourceLocation FIRE_CLAY_TEXTURE = new ResourceLocation(MOD_ID, "textures/gui/knapping/clay_button_fire.png");
    private static final ResourceLocation LEATHER_TEXTURE = new ResourceLocation(MOD_ID, "textures/gui/knapping/leather_button.png");

    // use this instead of player.openGui() -> avoids magic numbers
    public static void openGui(World world, BlockPos pos, EntityPlayer player, Type type)
    {
        player.openGui(TerraFirmaCraft.getInstance(), type.ordinal(), world, pos.func_177958_n(), pos.func_177956_o(), pos.func_177952_p());
    }

    // Only use this for things that don't need a BlockPos to identify TE's!!!
    public static void openGui(World world, EntityPlayer player, Type type)
    {
        player.openGui(TerraFirmaCraft.getInstance(), type.ordinal(), world, 0, 0, 0);
    }

    @Override
    @Nullable
    public Container getServerGuiElement(int ID, EntityPlayer player, World world, int x, int y, int z)
    {
        BlockPos pos = new BlockPos(x, y, z);
        ItemStack stack = player.func_184614_ca();
        Type type = Type.valueOf(ID);
        switch (type)
        {
            case NEST_BOX:
                TENestBox teNestBox = Helpers.getTE(world, pos, TENestBox.class);
                return teNestBox == null ? null : new ContainerNestBox(player.field_71071_by, teNestBox);
            case LOG_PILE:
                TELogPile teLogPile = Helpers.getTE(world, pos, TELogPile.class);
                return teLogPile == null ? null : new ContainerLogPile(player.field_71071_by, teLogPile);
            case SMALL_VESSEL:
                return new ContainerSmallVessel(player.field_71071_by, stack.func_77973_b() instanceof ItemSmallVessel ? stack : player.func_184592_cb());
            case SMALL_VESSEL_LIQUID:
                return new ContainerLiquidTransfer(player.field_71071_by, stack.func_77973_b() instanceof ItemSmallVessel ? stack : player.func_184592_cb());
            case MOLD:
                return new ContainerLiquidTransfer(player.field_71071_by, stack.func_77973_b() instanceof ItemMold ? stack : player.func_184592_cb());
            case FIRE_PIT:
                return new ContainerFirePit(player.field_71071_by, Helpers.getTE(world, pos, TEFirePit.class));
            case BARREL:
                return new ContainerBarrel(player.field_71071_by, Helpers.getTE(world, pos, TEBarrel.class));
            case CHARCOAL_FORGE:
                return new ContainerCharcoalForge(player.field_71071_by, Helpers.getTE(world, pos, TECharcoalForge.class));
            case ANVIL:
                return new ContainerAnvilTFC(player.field_71071_by, Helpers.getTE(world, pos, TEAnvilTFC.class));
            case ANVIL_PLAN:
                return new ContainerAnvilPlan(player.field_71071_by, Helpers.getTE(world, pos, TEAnvilTFC.class));
            case KNAPPING_STONE:
                return new ContainerKnapping(KnappingRecipe.Type.STONE, player.field_71071_by, stack.func_77973_b() instanceof ItemRock ? stack : player.func_184592_cb());
            case KNAPPING_CLAY:
                return new ContainerKnapping(KnappingRecipe.Type.CLAY, player.field_71071_by, stack.func_77973_b() == Items.field_151119_aD ? stack : player.func_184592_cb());
            case KNAPPING_LEATHER:
                return new ContainerKnapping(KnappingRecipe.Type.LEATHER, player.field_71071_by, stack.func_77973_b() == Items.field_151116_aA ? stack : player.func_184592_cb());
            case KNAPPING_FIRE_CLAY:
                return new ContainerKnapping(KnappingRecipe.Type.FIRE_CLAY, player.field_71071_by, stack.func_77973_b() == ItemsTFC.FIRE_CLAY ? stack : player.func_184592_cb());
            case QUERN:
                return new ContainerQuern(player.field_71071_by, Helpers.getTE(world, pos, TEQuern.class));
            case CRUCIBLE:
                return new ContainerCrucible(player.field_71071_by, Helpers.getTE(world, pos, TECrucible.class));
            case LARGE_VESSEL:
                return new ContainerLargeVessel(player.field_71071_by, Helpers.getTE(world, pos, TELargeVessel.class));
            case CALENDAR:
            case SKILLS:
            case NUTRITION:
                return new ContainerSimple(player.field_71071_by);
            case BLAST_FURNACE:
                return new ContainerBlastFurnace(player.field_71071_by, Helpers.getTE(world, pos, TEBlastFurnace.class));
            case CRAFTING:
                return new ContainerInventoryCrafting(player.field_71071_by, player.field_70170_p);
            default:
                return null;
        }
    }

    @Override
    @Nullable
    public Object getClientGuiElement(int ID, EntityPlayer player, World world, int x, int y, int z)
    {
        Container container = getServerGuiElement(ID, player, world, x, y, z);
        Type type = Type.valueOf(ID);
        BlockPos pos = new BlockPos(x, y, z);
        switch (type)
        {
            case NEST_BOX:
                return new GuiContainerTFC(container, player.field_71071_by, SMALL_INVENTORY_BACKGROUND);
            case LOG_PILE:
                return new GuiContainerTFC(container, player.field_71071_by, SMALL_INVENTORY_BACKGROUND);
            case SMALL_VESSEL:
                return new GuiContainerTFC(container, player.field_71071_by, SMALL_INVENTORY_BACKGROUND);
            case SMALL_VESSEL_LIQUID:
                return new GuiLiquidTransfer(container, player, player.func_184614_ca().func_77973_b() instanceof ItemSmallVessel);
            case MOLD:
                return new GuiLiquidTransfer(container, player, player.func_184614_ca().func_77973_b() instanceof ItemMold);
            case FIRE_PIT:
                return new GuiFirePit(container, player.field_71071_by, Helpers.getTE(world, pos, TEFirePit.class));
            case BARREL:
                return new GuiBarrel(container, player.field_71071_by, Helpers.getTE(world, pos, TEBarrel.class), world.func_180495_p(new BlockPos(x, y, z)).func_177230_c().func_149739_a());
            case CHARCOAL_FORGE:
                return new GuiCharcoalForge(container, player.field_71071_by, Helpers.getTE(world, pos, TECharcoalForge.class));
            case ANVIL:
                return new GuiAnvilTFC(container, player.field_71071_by, Helpers.getTE(world, pos, TEAnvilTFC.class));
            case ANVIL_PLAN:
                return new GuiAnvilPlan(container, player.field_71071_by, Helpers.getTE(world, pos, TEAnvilTFC.class));
            case KNAPPING_STONE:
                ItemStack stack = player.func_184614_ca();
                Rock rock = stack.func_77973_b() instanceof IRockObject ? ((IRockObject) stack.func_77973_b()).getRock(stack) :
                    ((IRockObject) player.func_184592_cb().func_77973_b()).getRock(player.func_184592_cb());
                //noinspection ConstantConditions
                return new GuiKnapping(container, player, KnappingRecipe.Type.STONE, rock.getTexture());
            case KNAPPING_CLAY:
                return new GuiKnapping(container, player, KnappingRecipe.Type.CLAY, CLAY_TEXTURE);
            case KNAPPING_LEATHER:
                return new GuiKnapping(container, player, KnappingRecipe.Type.LEATHER, LEATHER_TEXTURE);
            case KNAPPING_FIRE_CLAY:
                return new GuiKnapping(container, player, KnappingRecipe.Type.FIRE_CLAY, FIRE_CLAY_TEXTURE);
            case QUERN:
                return new GuiQuern(container, player.field_71071_by, Helpers.getTE(world, pos, TEQuern.class), world.func_180495_p(new BlockPos(x, y, z)).func_177230_c().func_149739_a());
            case CRUCIBLE:
                return new GuiCrucible(container, player.field_71071_by, Helpers.getTE(world, pos, TECrucible.class));
            case LARGE_VESSEL:
                return new GuiLargeVessel(container, player.field_71071_by, Helpers.getTE(world, pos, TELargeVessel.class), world.func_180495_p(new BlockPos(x, y, z)).func_177230_c().func_149739_a());
            case CALENDAR:
                return new GuiCalendar(container, player.field_71071_by);
            case NUTRITION:
                return new GuiNutrition(container, player.field_71071_by);
            case SKILLS:
                return new GuiSkills(container, player.field_71071_by);
            case BLAST_FURNACE:
                return new GuiBlastFurnace(container, player.field_71071_by, Helpers.getTE(world, pos, TEBlastFurnace.class));
            case CRAFTING:
                return new GuiInventoryCrafting(container);
            default:
                return null;
        }
    }

    public enum Type
    {
        NEST_BOX,
        LOG_PILE,
        SMALL_VESSEL,
        SMALL_VESSEL_LIQUID,
        MOLD,
        FIRE_PIT,
        BARREL,
        QUERN,
        KNAPPING_STONE,
        KNAPPING_CLAY,
        KNAPPING_FIRE_CLAY,
        KNAPPING_LEATHER,
        CHARCOAL_FORGE,
        ANVIL,
        ANVIL_PLAN,
        CRUCIBLE,
        BLAST_FURNACE,
        LARGE_VESSEL,
        CALENDAR,
        NUTRITION,
        SKILLS,
        INVENTORY, // This is special, it is used by GuiButtonPlayerInventoryTab to signal to open the vanilla inventory
        CRAFTING, // In-inventory 3x3 crafting grid
        NULL; // This is special, it is a non-null null.

        private static Type[] values = values();

        @Nonnull
        public static Type valueOf(int id)
        {
            return id < 0 || id >= values.length ? NULL : values[id];
        }
    }
}
