/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.objects.blocks.plants;

import java.util.HashMap;
import java.util.Map;
import java.util.Random;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;

import net.minecraft.block.Block;
import net.minecraft.block.BlockBush;
import net.minecraft.block.SoundType;
import net.minecraft.block.properties.PropertyInteger;
import net.minecraft.block.state.BlockStateContainer;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.Blocks;
import net.minecraft.init.Items;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.math.AxisAlignedBB;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.EnumSkyBlock;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.World;
import net.minecraftforge.common.EnumPlantType;

import net.dries007.tfc.ConfigTFC;
import net.dries007.tfc.api.capability.size.IItemSize;
import net.dries007.tfc.api.capability.size.Size;
import net.dries007.tfc.api.capability.size.Weight;
import net.dries007.tfc.api.types.Plant;
import net.dries007.tfc.objects.blocks.BlocksTFC;
import net.dries007.tfc.util.OreDictionaryHelper;
import net.dries007.tfc.util.calendar.CalendarTFC;
import net.dries007.tfc.util.calendar.ICalendar;
import net.dries007.tfc.util.calendar.Month;
import net.dries007.tfc.world.classic.ClimateTFC;
import net.dries007.tfc.world.classic.chunkdata.ChunkDataTFC;

@ParametersAreNonnullByDefault
public class BlockPlantTFC extends BlockBush implements IItemSize
{
    public static final PropertyInteger AGE = PropertyInteger.func_177719_a("age", 0, 3);
    /*
     * Time of day, used for rendering plants that bloom at different times
     * 0 = midnight-dawn
     * 1 = dawn-noon
     * 2 = noon-dusk
     * 3 = dusk-midnight
     */
    public final static PropertyInteger DAYPERIOD = PropertyInteger.func_177719_a("dayperiod", 0, 3);
    private static final AxisAlignedBB PLANT_AABB = new AxisAlignedBB(0.125D, 0.0D, 0.125D, 0.875D, 1.0D, 0.875D);
    private static final Map<Plant, BlockPlantTFC> MAP = new HashMap<>();

    public static BlockPlantTFC get(Plant plant)
    {
        return MAP.get(plant);
    }

    /* Growth Stage of the plant, tied to the month of year */
    public final PropertyInteger growthStageProperty;
    protected final Plant plant;
    protected final BlockStateContainer blockState;

    public BlockPlantTFC(Plant plant)
    {
        super(plant.getMaterial());
        if (MAP.put(plant, this) != null) throw new IllegalStateException("There can only be one.");

        plant.getOreDictName().ifPresent(name -> OreDictionaryHelper.register(this, name));

        this.plant = plant;
        this.growthStageProperty = PropertyInteger.func_177719_a("stage", 0, plant.getNumStages());
        this.func_149675_a(true);
        func_149672_a(SoundType.field_185850_c);
        func_149711_c(0.0F);
        Blocks.field_150480_ab.func_180686_a(this, 5, 20);
        blockState = this.createPlantBlockState();
        this.func_180632_j(this.blockState.func_177621_b());
    }

    @SuppressWarnings("deprecation")
    @Override
    @Nonnull
    public IBlockState func_176203_a(int meta)
    {
        return this.func_176223_P().func_177226_a(AGE, meta);
    }

    @Override
    public int func_176201_c(IBlockState state)
    {
        return state.func_177229_b(AGE);
    }

    @SuppressWarnings("deprecation")
    @Override
    @Nonnull
    public IBlockState func_176221_a(IBlockState state, IBlockAccess worldIn, BlockPos pos)
    {
        return state.func_177226_a(DAYPERIOD, getDayPeriod()).func_177226_a(growthStageProperty, plant.getStageForMonth());
    }

    @Override
    public boolean func_176200_f(IBlockAccess worldIn, BlockPos pos)
    {
        return true;
    }

    @Override
    public void func_180645_a(World worldIn, BlockPos pos, IBlockState state, Random random)
    {
        if (!worldIn.func_175697_a(pos, 1)) return;
        Month currentMonth = CalendarTFC.CALENDAR_TIME.getMonthOfYear();
        int currentStage = state.func_177229_b(growthStageProperty);
        int expectedStage = plant.getStageForMonth(currentMonth);
        int currentTime = state.func_177229_b(DAYPERIOD);
        int expectedTime = getDayPeriod();

        if (currentTime != expectedTime)
        {
            worldIn.func_175656_a(pos, state.func_177226_a(DAYPERIOD, expectedTime).func_177226_a(growthStageProperty, currentStage));
        }
        if (currentStage != expectedStage && random.nextDouble() < 0.5)
        {
            worldIn.func_175656_a(pos, state.func_177226_a(DAYPERIOD, expectedTime).func_177226_a(growthStageProperty, expectedStage));
        }

        this.func_180650_b(worldIn, pos, state, random);
    }

    @Override
    public int func_149738_a(World worldIn)
    {
        return 10;
    }

    @Override
    public void func_176213_c(World world, BlockPos pos, IBlockState state)
    {
        world.func_175656_a(pos, state.func_177226_a(DAYPERIOD, getDayPeriod()).func_177226_a(growthStageProperty, plant.getStageForMonth()));
        func_176475_e(world, pos, state);
    }

    @Override
    @Nonnull
    public Item func_180660_a(IBlockState state, Random rand, int fortune)
    {
        if (!plant.getOreDictName().isPresent()) return Items.field_190931_a;
        return Item.func_150898_a(this);
    }

    @Override
    public void func_180634_a(World worldIn, BlockPos pos, IBlockState state, Entity entityIn)
    {
        double movementMod = plant.getMovementMod();
        //Entity X/Z motion is reduced by plants.
        // todo: make this specific to the plant type (i.e. smaller grasses affect you less than larger ones)
        entityIn.field_70159_w *= movementMod;
        entityIn.field_70179_y *= movementMod;
    }

    @Override
    public void func_180657_a(World worldIn, EntityPlayer player, BlockPos pos, IBlockState state, @Nullable TileEntity te, ItemStack stack)
    {
        if (!plant.getOreDictName().isPresent() && !worldIn.field_72995_K && (stack.func_77973_b().getHarvestLevel(stack, "knife", player, state) != -1 || stack.func_77973_b().getHarvestLevel(stack, "scythe", player, state) != -1) && plant.getPlantType() != Plant.PlantType.SHORT_GRASS && plant.getPlantType() != Plant.PlantType.TALL_GRASS)
        {
            func_180635_a(worldIn, pos, new ItemStack(this, 1));
        }
        super.func_180657_a(worldIn, player, pos, state, te, stack);
    }

    @Override
    public void func_180633_a(World worldIn, BlockPos pos, IBlockState state, EntityLivingBase placer, ItemStack stack)
    {
        if (!func_180671_f(worldIn, pos, state) && placer instanceof EntityPlayer)
        {
            if (!((EntityPlayer) placer).func_184812_l_() && !plant.getOreDictName().isPresent())
            {
                func_180635_a(worldIn, pos, new ItemStack(this));
            }
        }
    }

    @Nonnull
    @Override
    public BlockStateContainer func_176194_O()
    {
        return this.blockState;
    }

    @Override
    @Nonnull
    public Block.EnumOffsetType func_176218_Q()
    {
        return Block.EnumOffsetType.XYZ;
    }

    public Plant getPlant()
    {
        return plant;
    }

    @Nonnull
    @Override
    public Size getSize(ItemStack stack)
    {
        return Size.SMALL;
    }

    @Nonnull
    @Override
    public Weight getWeight(ItemStack stack)
    {
        return Weight.LIGHT;
    }

    public double getGrowthRate(World world, BlockPos pos)
    {
        if (world.func_175727_C(pos)) return ConfigTFC.GENERAL.plantGrowthRate * 5d;
        else return ConfigTFC.GENERAL.plantGrowthRate;
    }

    @Override
    public boolean func_176196_c(World worldIn, BlockPos pos)
    {
        IBlockState soil = worldIn.func_180495_p(pos.func_177977_b());
        return worldIn.func_180495_p(pos).func_177230_c().func_176200_f(worldIn, pos) && worldIn.func_180495_p(pos).func_177230_c() != this && soil.func_177230_c().canSustainPlant(soil, worldIn, pos.func_177977_b(), net.minecraft.util.EnumFacing.UP, this);
    }

    @Override
    protected boolean func_185514_i(IBlockState state)
    {
        if (plant.getIsClayMarking()) return BlocksTFC.isClay(state) || isValidSoil(state);
        else return isValidSoil(state);
    }

    @Override
    public void func_180650_b(World worldIn, BlockPos pos, IBlockState state, Random rand)
    {
        if (!worldIn.func_175697_a(pos, 1)) return;

        if (plant.isValidGrowthTemp(ClimateTFC.getHeightAdjustedTemp(worldIn, pos)) && plant.isValidSunlight(Math.subtractExact(worldIn.func_175642_b(EnumSkyBlock.SKY, pos), worldIn.func_175657_ab())))
        {
            int j = state.func_177229_b(AGE);

            if (rand.nextDouble() < getGrowthRate(worldIn, pos) && net.minecraftforge.common.ForgeHooks.onCropsGrowPre(worldIn, pos.func_177984_a(), state, true))
            {
                if (j < 3)
                {
                    worldIn.func_175656_a(pos, state.func_177226_a(AGE, j + 1));
                }
                net.minecraftforge.common.ForgeHooks.onCropsGrowPost(worldIn, pos, state, worldIn.func_180495_p(pos));
            }
        }
        else if (!plant.isValidGrowthTemp(ClimateTFC.getHeightAdjustedTemp(worldIn, pos)) || !plant.isValidSunlight(worldIn.func_175642_b(EnumSkyBlock.SKY, pos)))
        {
            int j = state.func_177229_b(AGE);

            if (rand.nextDouble() < getGrowthRate(worldIn, pos) && net.minecraftforge.common.ForgeHooks.onCropsGrowPre(worldIn, pos, state, true))
            {
                if (j > 0)
                {
                    worldIn.func_175656_a(pos, state.func_177226_a(AGE, j - 1));
                }
                net.minecraftforge.common.ForgeHooks.onCropsGrowPost(worldIn, pos, state, worldIn.func_180495_p(pos));
            }
        }

        func_176475_e(worldIn, pos, state);
    }

    @Override
    public boolean func_180671_f(World worldIn, BlockPos pos, IBlockState state)
    {
        IBlockState soil = worldIn.func_180495_p(pos.func_177977_b());
        if (state.func_177230_c() == this)
        {
            return soil.func_177230_c().canSustainPlant(soil, worldIn, pos.func_177977_b(), net.minecraft.util.EnumFacing.UP, this) && plant.isValidTemp(ClimateTFC.getHeightAdjustedTemp(worldIn, pos)) && plant.isValidRain(ChunkDataTFC.getRainfall(worldIn, pos));
        }
        return this.func_185514_i(soil);
    }

    @SuppressWarnings("deprecation")
    @Override
    @Nonnull
    public AxisAlignedBB func_185496_a(IBlockState state, IBlockAccess source, BlockPos pos)
    {
        return PLANT_AABB.func_191194_a(state.func_191059_e(source, pos));
    }

    @SuppressWarnings("deprecation")
    @Nullable
    @Override
    public AxisAlignedBB func_180646_a(IBlockState blockState, IBlockAccess worldIn, BlockPos pos)
    {
        if (plant.getMovementMod() == 0.0D)
        {
            return blockState.func_185900_c(worldIn, pos);
        }
        else
        {
            return field_185506_k;
        }
    }

    @Override
    @Nonnull
    public EnumPlantType getPlantType(net.minecraft.world.IBlockAccess world, BlockPos pos)
    {
        switch (plant.getPlantType())
        {
            case CACTUS:
            case DESERT:
            case DESERT_TALL_PLANT:
                return EnumPlantType.Desert;
            case FLOATING:
            case FLOATING_SEA:
                return EnumPlantType.Water;
            case MUSHROOM:
                return EnumPlantType.Cave;
            default:
                return EnumPlantType.Plains;
        }
    }

    @Nonnull
    public Plant.EnumPlantTypeTFC getPlantTypeTFC()
    {
        return plant.getEnumPlantTypeTFC();
    }

    @Nonnull
    protected BlockStateContainer createPlantBlockState()
    {
        return new BlockStateContainer(this, growthStageProperty, DAYPERIOD, AGE);
    }

    int getDayPeriod()
    {
        return Math.floorDiv(CalendarTFC.CALENDAR_TIME.getHourOfDay(), ICalendar.HOURS_IN_DAY / 4);
    }

    private boolean isValidSoil(IBlockState state)
    {
        switch (plant.getPlantType())
        {
            case CACTUS:
            case DESERT:
            case DESERT_TALL_PLANT:
                return BlocksTFC.isSand(state);
            case DRY:
            case DRY_TALL_PLANT:
                return BlocksTFC.isSand(state) || BlocksTFC.isDryGrass(state);
            case REED:
            case REED_SEA:
            case TALL_REED:
            case TALL_REED_SEA:
                return BlocksTFC.isSand(state) || BlocksTFC.isSoil(state);
            case WATER:
            case TALL_WATER:
            case EMERGENT_TALL_WATER:
                return BlocksTFC.isSoilOrGravel(state);
            case WATER_SEA:
            case TALL_WATER_SEA:
            case EMERGENT_TALL_WATER_SEA:
                return BlocksTFC.isSand(state) || BlocksTFC.isSoilOrGravel(state);
            default:
                return BlocksTFC.isSoil(state);
        }
    }
}
