/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.objects.entity.animal;

import java.util.List;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;

import net.minecraft.entity.EntityAgeable;
import net.minecraft.entity.passive.EntityAnimal;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.SoundEvents;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.network.datasync.DataParameter;
import net.minecraft.network.datasync.DataSerializers;
import net.minecraft.network.datasync.EntityDataManager;
import net.minecraft.util.EnumHand;
import net.minecraft.util.SoundCategory;
import net.minecraft.util.text.TextComponentTranslation;
import net.minecraft.world.World;

import net.dries007.tfc.util.OreDictionaryHelper;
import net.dries007.tfc.util.calendar.CalendarTFC;
import net.dries007.tfc.util.calendar.ICalendar;

import static net.dries007.tfc.api.util.TFCConstants.MOD_ID;

@ParametersAreNonnullByDefault
public abstract class EntityAnimalTFC extends EntityAnimal
{
    private static final long DEFAULT_TICKS_COOLDOWN_MATING = ICalendar.TICKS_IN_HOUR * 2;
    //Values that has a visual effect on client
    private static final DataParameter<Boolean> GENDER = EntityDataManager.func_187226_a(EntityAnimalTFC.class, DataSerializers.field_187198_h);
    private static final DataParameter<Integer> BIRTHDAY = EntityDataManager.func_187226_a(EntityAnimalTFC.class, DataSerializers.field_187192_b);
    private static final DataParameter<Float> FAMILIARITY = EntityDataManager.func_187226_a(EntityAnimalTFC.class, DataSerializers.field_187193_c);

    private long lastFed; //Last time(in days) this entity was fed
    private long lastFDecay; //Last time(in days) this entity's familiarity had decayed

    private boolean fertilized; //Is this female fertilized? (in oviparous, the egg laying is fertilized, for mammals this is pregnancy)
    private long matingTime; //The last time(in ticks) this male tried fertilizing females

    public EntityAnimalTFC(World worldIn, Gender gender, int birthDay)
    {
        super(worldIn);
        this.setGender(gender);
        this.setBirthDay(birthDay);
        this.setFamiliarity(0);
        this.func_70873_a(0); //We don't use this
        this.lastFed = -1;
        this.matingTime = -1;
        this.lastFDecay = CalendarTFC.PLAYER_TIME.getTotalDays();
        this.fertilized = false;
    }

    @SuppressWarnings("unused")
    public EntityAnimalTFC(World worldIn)
    {
        super(worldIn);
    }

    public Gender getGender()
    {
        return Gender.fromBool(this.field_70180_af.func_187225_a(GENDER));
    }

    public void setGender(Gender gender)
    {
        this.field_70180_af.func_187227_b(GENDER, gender.toBool());
    }

    public int getBirthDay()
    {
        return this.field_70180_af.func_187225_a(BIRTHDAY);
    }

    public void setBirthDay(int value)
    {
        this.field_70180_af.func_187227_b(BIRTHDAY, value);
    }

    public float getFamiliarity()
    {
        return this.field_70180_af.func_187225_a(FAMILIARITY);
    }

    public void setFamiliarity(float value)
    {
        if (value < 0f) value = 0f;
        if (value > 1f) value = 1f;
        this.field_70180_af.func_187227_b(FAMILIARITY, value);
    }

    public boolean getIsFedToday()
    {
        return this.lastFed == CalendarTFC.PLAYER_TIME.getTotalDays();
    }

    public boolean isFertilized() { return this.fertilized; }

    public void setFertilized(boolean value)
    {
        this.fertilized = value;
    }

    @Override
    public void func_70636_d()
    {
        super.func_70636_d();
        if (!this.field_70170_p.field_72995_K)
        {
            //Is it time to decay familiarity?
            if (this.lastFDecay < CalendarTFC.PLAYER_TIME.getTotalDays())
            {
                float familiarity = getFamiliarity();
                //If this entity was never fed(eg: new born, wild)
                //or wasn't fed yesterday(this is the starting of the second day)
                if (familiarity < 0.3f && (this.lastFed == -1 || this.lastFed - 1 < CalendarTFC.PLAYER_TIME.getTotalDays()))
                {
                    familiarity -= 0.02 * (CalendarTFC.PLAYER_TIME.getTotalDays() - this.lastFDecay);
                    this.lastFDecay = CalendarTFC.PLAYER_TIME.getTotalDays();
                    if (familiarity < 0) familiarity = 0f;
                    this.setFamiliarity(familiarity);
                }
            }
            if (this.getGender() == Gender.MALE && this.isReadyToMate())
            {
                this.matingTime = CalendarTFC.PLAYER_TIME.getTicks();
                if (findFemaleMate())
                {
                    this.func_146082_f(null);
                }
            }
        }
    }

    @Override
    public void func_70014_b(NBTTagCompound nbt)
    {
        super.func_70014_b(nbt);
        nbt.func_74757_a("gender", getGender().toBool());
        nbt.func_74768_a("birth", getBirthDay());
        nbt.func_74772_a("fed", lastFed);
        nbt.func_74772_a("decay", lastFDecay);
        nbt.func_74757_a("fertilized", this.fertilized);
        nbt.func_74772_a("mating", matingTime);
        nbt.func_74776_a("familiarity", getFamiliarity());
    }

    @Override
    public void func_70037_a(NBTTagCompound nbt)
    {
        super.func_70037_a(nbt);
        this.setGender(Gender.fromBool(nbt.func_74767_n("gender")));
        this.setBirthDay(nbt.func_74762_e("birth"));
        this.lastFed = nbt.func_74763_f("fed");
        this.lastFDecay = nbt.func_74763_f("decay");
        this.matingTime = nbt.func_74763_f("mating");
        this.fertilized = nbt.func_74767_n("fertilized");
        this.setFamiliarity(nbt.func_74760_g("familiarity"));

    }

    @Override
    public boolean func_70877_b(ItemStack stack)
    {
        return OreDictionaryHelper.doesStackMatchOre(stack, "grain");
    }

    @Override
    public boolean func_184645_a(EntityPlayer player, EnumHand hand)
    {
        ItemStack itemstack = player.func_184586_b(hand);

        if (!itemstack.func_190926_b())
        {
            if (this.func_70877_b(itemstack) && player.func_70093_af())
            {
                if (this.canFeed())
                {
                    if (!this.field_70170_p.field_72995_K)
                    {
                        lastFed = CalendarTFC.PLAYER_TIME.getTotalDays();
                        lastFDecay = lastFed; //No decay needed
                        this.func_175505_a(player, itemstack);
                        this.setFamiliarity(this.getFamiliarity() + 0.06f);
                        field_70170_p.func_184133_a(null, this.func_180425_c(), SoundEvents.field_187739_dZ, SoundCategory.AMBIENT, 1.0F, 1.0F);
                    }
                    return true;
                }
                else
                {
                    if (!this.field_70170_p.field_72995_K)
                    {
                        //Show tooltips
                        if (this.isFertilized() && this instanceof EntityAnimalMammal)
                        {
                            player.func_145747_a(new TextComponentTranslation(MOD_ID + ".tooltip.animal.pregnant"));
                        }
                    }
                }
            }
        }
        return false;
    }

    @Override
    public boolean func_70878_b(EntityAnimal otherAnimal)
    {
        if (otherAnimal.getClass() != this.getClass()) return false;
        EntityAnimalTFC other = (EntityAnimalTFC) otherAnimal;
        return this.getGender() != other.getGender() && this.func_70880_s() && other.func_70880_s();
    }

    /**
     * Event, used by children of this class, to do things on fertilization of females
     */
    public void onFertilized(EntityAnimalTFC male)
    {
    }

    @Nullable
    @Override
    public EntityAgeable func_90011_a(@Nonnull EntityAgeable other)
    {
        if (this.getGender() == Gender.FEMALE)
        {
            this.fertilized = true;
            this.func_70875_t();
            this.onFertilized((EntityAnimalTFC) other);
        }
        return null;
    }

    @Override
    protected void func_70088_a()
    {
        super.func_70088_a();
        func_184212_Q().func_187214_a(GENDER, true);
        func_184212_Q().func_187214_a(BIRTHDAY, 0);
        func_184212_Q().func_187214_a(FAMILIARITY, 0f);
    }

    @Override
    public boolean func_70631_g_()
    {
        return this.getAge() == Age.CHILD;
    }

    @Override
    public void func_98054_a(boolean child)
    {
        float ageScale = 1 / (2.0F - getPercentToAdulthood());
        this.func_98055_j(ageScale);
    }

    /**
     * Used by models renderer to scale the size of the animal
     *
     * @return float value between 0(birthday) to 1(full grown adult)
     */
    public abstract float getPercentToAdulthood();

    /**
     * Get this entity age, based on birth
     *
     * @return the Age enum of this entity
     */
    public abstract Age getAge();

    /**
     * Find and charms a near female animal of this animal
     * Used by males to try mating with females
     *
     * @return true if found and charmed a female
     */
    protected boolean findFemaleMate()
    {
        List<EntityAnimalTFC> list = this.field_70170_p.func_72872_a(this.getClass(), this.func_174813_aQ().func_186662_g(8.0D));
        for (EntityAnimalTFC ent : list)
        {
            if (ent.getGender() == Gender.FEMALE && !ent.func_70880_s() && ent.isReadyToMate())
            {
                ent.func_146082_f(null);
                return true;
            }
        }
        return false;
    }

    /**
     * Check if this animal is ready to mate
     *
     * @return true if ready
     */
    protected boolean isReadyToMate()
    {
        if (this.getAge() != Age.ADULT || this.getFamiliarity() < 0.3f || this.isFertilized() || !this.getIsFedToday())
            return false;
        return this.matingTime == -1 || this.matingTime + getCooldownMating() <= CalendarTFC.PLAYER_TIME.getTicks();
    }

    /**
     * The number of ticks this animal needs to rest before trying to mate again
     *
     * @return ticks needed to fully recover
     */
    protected long getCooldownMating()
    {
        return DEFAULT_TICKS_COOLDOWN_MATING;
    }

    private boolean canFeed()
    {
        if (lastFed == -1) return true;
        return lastFed < CalendarTFC.PLAYER_TIME.getTotalDays();
    }

    public enum Age
    {
        CHILD, ADULT, OLD
    }

    public enum Gender
    {
        MALE, FEMALE;

        public static Gender fromBool(boolean value)
        {
            return value ? MALE : FEMALE;
        }

        public boolean toBool()
        {
            return this == MALE;
        }
    }
}
